/**
 * 要使得 babel-register 和 typescript-register 同时工作(先经过typescript编译，再由babel编译)
 * 必须满足:①babel-register>=7.0.0-beta.37，②两者依赖的source-map-support,和项目依赖的要全部一致
 */
const path = require('path');
const fs = require('fs');
const vm = require('vm');
const pirates = require('pirates');
const chalk = require('chalk');
const _ = require('lodash');
const sourceMapConverter = require('convert-source-map');
const sourceMapSupport = require('source-map-support');
const debug = require('./utils/debugPlus')('core');
const babelCore = require('@babel/core');
babelCore.DEFAULT_EXTENSIONS = null; //阻止babel-register初始注册，https://github.com/babel/babel/blob/master/packages/babel-register/src/node.js#L96
const babelRegister = require('@babel/register');
const tsNode = require('ts-node');
const tsConfigTools = require('tsconfig');
const boxen = require('boxen');
const figlet = require('figlet');
const cosmiconfig = require('cosmiconfig');
const json5 = require('json5');
// const prettyError = require('pretty-error');
const explorer = cosmiconfig('tbRegister', {
  cache: false,
  packageProp: "tb-register",
  loaders: {
    ".json": function json5Loader(fileName, content) {
      return json5.parse(content);
    }
  }
});
const rcConfig = _.get(explorer.searchSync(), 'config', {});

const utils = require('./utils/index');
const {
  esModToCommon,
  mapSourcePosition,
  prettyError,
  getRawCode,
  getUserConfig,
  getHelpfulFrame,
  getCodeFrame,
  openEditor,
  genModuleResolverConfig
} = utils;
const babelModuleResolver = esModToCommon(require('babel-plugin-module-resolver'));
const babelSyntaxDynamicImport = esModToCommon(require('babel-plugin-syntax-dynamic-import'));
const babelDynamicImportNode = esModToCommon(require('babel-plugin-dynamic-import-node'));
const babelPresetEnv = esModToCommon(require('babel-preset-env'));
const babelSyntaxObjectRestspread = esModToCommon(require('babel-plugin-syntax-object-rest-spread')); //只识别语法不作转换（node>=8.3.0已支持）

const KEY_REAL_STACK = Symbol.for('realStack');
const KEY_ORIGIN_STACK = Symbol.for('originStack');
const KEY_ERROR_POSITIONS = Symbol.for('errorPositions');
const KEY_PREPARESTACKTRACE = Symbol.for('prepareStackTrace');
const CWD = process.cwd();

const RE_INLINE_SOURCEMAP = /\/\/\# sourceMappingURL\=(.+)/i; //匹配代码(经过babel,typescript处理后)里的内联 sourcemap
const RE_NODE_MODULE = /node_modules/i;
const RE_EXTRACT_INFO_FROM_STANDARD_STACK = /\((.+?):(\d+):(\d+)\)/; //从栈帧中提取文件路径($1)、行号($2)和列号($3)
const RE_EXTRACT_INFO_FROM_TS_COMPILE_ERROR = /(\S+?\.tsx?)\s+\((\d+),(\d+)\)\:\s*(.*)/i; //从ts编译错误的信息中提取文件路径($1)、行号($2)、列号($3)、描述$(4)
//从vm.js(vm.runInThisContext)抛出的错误(SyntaxError)信息(error.stack)中提取文件路径($1)、行号($2)、列号($3.length)、错误描述($5)
const RE_EXTRACE_INFO_FROM_VM_ERROR_MESSAGE = /^([^\n]+):(\d+)\n(?:.+?\n)( *)(\^+) *\n(\n?.+?\n)/;


let installed = false;

function register(config = {}) {
  debugger;
  if (installed) throw Error('Prohibition of multiple registrations');
  const getConfig = getUserConfig.bind(null, Object.assign({}, rcConfig, config));
  const rootPath = getConfig('rootPath', CWD);
  const rootPathAlias = getConfig('rootPathAlias', 'local://__ROOT__');
  const sourceMapRootPathAlias = getConfig('sourceMapRootPathAlias', 'local://__SOURCE_MAPS__');
  const useBabel = !!getConfig('babel', true);
  const useTs = !!getConfig('typescript', true);
  const usePrettyError = !!getConfig('prettyError', true);
  const useOpenEditor = !!getConfig('openEditor', true);
  // console.log('32',usePrettyError)
  const {
    mockFilePath,
    retrieveFilePath,
    repairSourceMap,
    repairStackFilename
  } = utils.make({
    rootPath,
    rootPathAlias,
    sourceMapRootPathAlias
  });
  let typescriptConfig = null;
  let defaultBabelModuleResolverConf = null;
  let babelConfig = null;
  let prettyConfig = null;
  let openEditorConfig = null;
  let oldErrorTracer, prettyErrorTracer;

  debug(chalk.blueBright(`useBabel:${useBabel}`));
  debug(chalk.blueBright(`useTs:${useTs}`));
  if (useTs) {
    typescriptConfig = getConfig('typescript', {});
    const {
      config: tsCwdConfig,
      path: tsConfigPath
    } = tsConfigTools.loadSync(CWD);
    defaultBabelModuleResolverConf = genModuleResolverConfig({
        ...tsCwdConfig,
        ...typescriptConfig
      },
      tsConfigPath
    );
    debug('TsConfig:%O', typescriptConfig);
  }
  if (useBabel) {
    const babelPresets = getConfig('babelPresets', [
      [
        babelPresetEnv,
        {
          targets: {
            node: 'current'
          }
        }
      ]
    ]);
    const babelModuleResolverConf = getConfig('babelModuleResolver', defaultBabelModuleResolverConf || genModuleResolverConfig());
    debug('babelModuleResolver:%O', babelModuleResolver);
    const babelPlugins = getConfig('babelPlugins', []);
    babelConfig = getConfig('babel', {
      cache: false, //@babel-register7.0.0-beta.37有BUG启用缓存时无法从缓存中读取sourcemap
      extensions: ['.ts', '.tsx'],
      babelrc: false,
      presets: babelPresets,
      plugins: [
        [babelModuleResolver, babelModuleResolverConf],
        babelSyntaxObjectRestspread,
        babelSyntaxDynamicImport,
        babelDynamicImportNode,
        ...babelPlugins
      ]
    });
    debug('babelConfig:%O', babelConfig);
  }

  if (usePrettyError) {
    const defaultPrettyConfig = {
      skipNodeFiles: getConfig('prettyErrorSkipNodeFiles', ['module.js']),
      skipPackage: getConfig('prettyErrorSkipPackage', []),
      alias: getConfig('prettyErrorAlias', [
        [CWD, '__ROOT__']
      ]),
      style: getConfig('prettyErrorStyle', prettyError.DEFAULT_STYLE)
    };

    prettyConfig = getConfig('prettyError', defaultPrettyConfig);
    prettyConfig = typeof prettyConfig !== 'object' ? defaultPrettyConfig : prettyConfig;
    debug('prettyConfig:%O', prettyConfig);
  }

  if (useOpenEditor) {
    openEditorConfig = getConfig('openEditor', openEditor.DEFAULT_CONFIG);
  }
  /**
   * 解决window下使用chrome devtool调试nodejs程序时无法读取sourcemap的问题
   * 参考{@link https://github.com/nodejs/node/issues/8369#issuecomment-321722973}这种解决方案只是临时的
   * node module.js源码{@link https://github.com/nodejs/node/blob/master/lib/module.js#L628}
   */
  vm._runInThisContext = vm.runInThisContext;
  vm.runInThisContext = function (code, opts) {
    try {
      return vm._runInThisContext(code, {
        ...opts,
        filename: mockFilePath(opts.filename)
      });
    } catch (err) {
      if (err.name === 'SyntaxError') {
        /**
         * SyntaxError是特殊情况 无法在Error.prepareStackTrace中获取有用的信息，所以在这里捕获&修改
         */
        err.stack = err.stack.replace(RE_EXTRACE_INFO_FROM_VM_ERROR_MESSAGE, function (matchStr, p1, p2, p3, p4, p5) {
          const filePath = retrieveFilePath(p1);
          const {
            source,
            line,
            column
          } = mapSourcePosition({
            source: filePath,
            line: p2,
            column: p3.length
          });
          err[KEY_ERROR_POSITIONS] = {
            filePath,
            lineNum: line,
            columnNum: column
          };
          return !usePrettyError ?
            matchStr.replace(p1, filePath) :
            p5 +
            '\n' +
            getCodeFrame(source, line, column, {
              linesAbove: 5,
              linesBelow: 3
            }) +
            '\n\n';
        });
      }
      throw err;
    }
  };

  useTs && tsNode.register(typescriptConfig);
  useBabel && babelRegister(babelConfig);
  sourceMapSupport.install({
    //这里重复初始化不影响，还防止了typescipt和babel都不启用时 Error.prepareStackTrace 没有值的问题
    handleUncaughtExceptions: false,
    environment: 'node'
  });

  /**
   * 修改由babel/typescript生成的inline sourcemap,控制sourcemap文件在chrome devtool中显示的名称
   */
  pirates.addHook(
    function fixSourcemap(code, filePath) {
      return code.replace(RE_INLINE_SOURCEMAP, function (str) {
        return repairSourceMap(str);
      });
    }, {
      exts: ['.js', '.ts', '.tsx'],
      matcher(filePath) {
        return !RE_NODE_MODULE.test(filePath);
      }
    }
  );

  oldErrorTracer = Error.prepareStackTrace;
  usePrettyError && prettyError(prettyConfig);

  prettyErrorTracer = Error.prepareStackTrace;

  // Error.stackTraceLimit = Infinity; // 设置堆栈跟踪收集的栈帧数量,默认10
  Error[KEY_PREPARESTACKTRACE] = diyPrepareStackTrace;

  function diyPrepareStackTrace(err, stack) {
    stack = repairStackFilename(stack);
    // if(stack.length>6){
    //   stack.splice(6);
    // }
    err[KEY_ORIGIN_STACK] = stack;
    if (err.name === 'SyntaxError') {
      if (/([\\\/])babylon\1/i.test(stack[0].getFileName())) {
        //判断是babel编译代码时抛出的SyntaxError;
        const result = err.message.match(/\s*(.+?):\s+(.+?)\s*\((\d+):(\d+)\)/);
        if (result) {
          const [, filePath, msg, lineNum, columnNum] = result;
          err.message = msg + '\n' + chalk.red(' ') + '\n' + getCodeFrame(filePath, lineNum, columnNum);
          err[KEY_ERROR_POSITIONS] = {
            filePath,
            lineNum,
            columnNum
          };
        }
      }
      return prettyErrorTracer(err, stack);
    }

    // typescript编译器抛出的错误
    if (usePrettyError && /Unable to compile TypeScript/.test(err.message)) {
      const infos = err.message.split(/\n/);
      const errorPositions = [];
      const newErrorFilesInfo = infos.slice(1).map(function (info) {
        const result = info.match(RE_EXTRACT_INFO_FROM_TS_COMPILE_ERROR);
        if (result) {
          const [, filePath, lineNum, columnNum, msg] = result;
          if (errorPositions.findIndex(v => v.filePath === filePath) === -1) {
            errorPositions.push({
              filePath,
              lineNum,
              columnNum
            });
          }
          const codeFrameContent = getCodeFrame(filePath, lineNum, columnNum, {
            highlightCode: true,
            showFilePath: false
          });
          return `${chalk.bgMagenta.whiteBright(`${filePath} (${lineNum},${columnNum})`)} - ${msg}\n${codeFrameContent}\n${chalk.red('\n')}`;
        }
        return info;
      });
      err[KEY_ERROR_POSITIONS] = errorPositions;
      err.message = `${infos[0]}\n${chalk.red(' ')}\n${newErrorFilesInfo.join('\n')}`;
      return prettyErrorTracer(err, stack);
    }

    const helpfulFrame = getHelpfulFrame(stack);
    if (helpfulFrame) {
      const filePath = helpfulFrame.getFileName();
      const lineNum = helpfulFrame.getLineNumber();
      const columnNum = helpfulFrame.getColumnNumber();
      err[KEY_ERROR_POSITIONS] = {
        filePath,
        lineNum,
        columnNum
      };
      if (usePrettyError) {
        const codeFrameContent = getCodeFrame(filePath, lineNum, columnNum, {
          linesAbove: 5,
          linesBelow: 3,
          highlightCode: true
        });
        return prettyErrorTracer(err, stack).replace(/\n/, `\n\n${codeFrameContent}\n`);
      }
    }
    return prettyErrorTracer(err, stack);
  }

  Object.defineProperty(Error, 'prepareStackTrace', {
    get() {
      return this[KEY_PREPARESTACKTRACE];
    },
    set(newVal) {
      if (debug.enabled) {
        const banner = boxen(
          [
            chalk.redBright(
              figlet.textSync('ERROR!', {
                horizontalLayout: 'full'
              })
            ),
            chalk.bgRed.whiteBright(`\n"Error.prepareStackTrace" is overwritten.`)
          ].join('\n'), {
            borderStyle: 'double',
            borderColor: 'red',
            align: 'center ',
            padding: 1,
            margin: 1
          }
        );
        const stackTraceText = chalk.redBright(
          Error()
          .realStack.slice(1)
          .map(v => `  at ${v}`)
          .join('\n')
        );
        console.log(`${banner}\n${stackTraceText}`);
        // return;
      }
      this[KEY_PREPARESTACKTRACE] = newVal;
    }
  });
  Object.defineProperty(Error.prototype, 'realStack', {
    get() {
      if (!this[KEY_ORIGIN_STACK]) this.stack;
      if (!this[KEY_ORIGIN_STACK]) return null;
      return (this[KEY_REAL_STACK] = this[KEY_REAL_STACK] || this[KEY_ORIGIN_STACK].map(v => sourceMapSupport.wrapCallSite(v)));
    }
  });
  useOpenEditor &&
    process.once('uncaughtException', function (err) {
      process.exitCode = 1;
      console.error(err.stack);
      if (useOpenEditor && err[KEY_ERROR_POSITIONS]) {
        openEditor([].concat(err[KEY_ERROR_POSITIONS]).map(v => `${v.filePath}:${v.lineNum}:${v.columnNum}`), {
          ...openEditorConfig,
          editor: 'code'
        });
      }
    });
  installed = true;
}
Object.defineProperty(register, 'installed', {
  get() {
    return installed;
  }
});
module.exports = register;